# Copyright 2009-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

"""
Tor curses monitoring application.

::

  nyx_interface - nyx interface singleton
  tor_controller - tor connection singleton
  cache - provides our application cache

  show_message - shows a message to the user
  input_prompt - prompts the user for text input
  init_controller - initializes our connection to tor
  expand_path - expands path with respect to our chroot
  chroot - provides the chroot path we reside within
  join - joins a series of strings up to a set length

  Cache - application cache
    |- write - provides a content where we can write to the cache
    |
    |- relay_nickname - provides the nickname of a relay
    +- relay_address - provides the address and orport of a relay

  CacheWriter - context in which we can write to the cache
    +- record_relay - caches information about a relay

  Interface - overall nyx interface
    |- get_page - page we're showing
    |- set_page - sets the page we're showing
    |- page_count - pages within our interface
    |
    |- header_panel - provides the header panel
    |- page_panels - provides panels on a page
    |
    |- is_paused - checks if the interface is paused
    |- set_paused - sets paused state
    |
    |- redraw - renders our content
    |- quit - quits our application
    +- halt - stops daemon panels
"""

import contextlib
import getpass
import os
import platform
import sys
import threading
import time

import stem
import stem.connection
import stem.control
import stem.util.conf
import stem.util.connection
import stem.util.log
import stem.util.system
import stem.util.tor_tools

SQLITE_UNAVAILABLE = """\
Python's sqlite3 module is unavailable. Unfortunately some platforms
don't bundle this with the interpreter. Please let us know at...

  https://trac.torproject.org/projects/tor/wiki/doc/nyx/bugs
"""

SQLITE_UNAVAILABLE_GENTOO = """\
Python's sqlite3 module is unavailable. For Gentoo please run
emerge dev-lang/python with USE=sqlite.
"""

SQLITE_UNAVAILABLE_FREEBSD = """\
Python's sqlite3 module is unavailable. Please run...

  sudo pkg install py%i%i-sqlite3
""" % sys.version_info[:2]

try:
  import sqlite3
except ImportError:
  if stem.util.system.is_gentoo():
    print(SQLITE_UNAVAILABLE_GENTOO)
  elif stem.util.system.is_bsd():
    print(SQLITE_UNAVAILABLE_FREEBSD)
  else:
    print(SQLITE_UNAVAILABLE)

  sys.exit(1)

__version__ = '2.1.0'
__release_date__ = 'January 12, 2019'
__author__ = 'Damian Johnson'
__contact__ = 'atagar@torproject.org'
__url__ = 'https://nyx.torproject.org/'
__license__ = 'GPLv3'

__all__ = [
  'arguments',
  'cache',
  'controller',
  'curses',
  'log',
  'menu',
  'panel',
  'popups',
  'starter',
  'tracker',
]


def conf_handler(key, value):
  if key == 'redraw_rate':
    return max(1, value)


CONFIG = stem.util.conf.config_dict('nyx', {
  'confirm_quit': True,
  'redraw_rate': 5,
  'show_graph': True,
  'show_log': True,
  'show_connections': True,
  'show_config': True,
  'show_torrc': True,
  'show_interpreter': True,
  'start_time': 0,
}, conf_handler)

NYX_INTERFACE = None
TOR_CONTROLLER = None
CACHE = None
CHROOT = None
BASE_DIR = os.path.sep.join(__file__.split(os.path.sep)[:-1])

# our address infrequently changes so we can cache it for a while

stem.control.CACHE_ADDRESS_FOR = 30

# disable trace level messages about cache hits

stem.control.LOG_CACHE_FETCHES = False

# don't parse NEWCONSENSUS events since we just care about the content

stem.response.events.PARSE_NEWCONSENSUS_EVENTS = False

# Duration for threads to pause when there's no work left to do. This is a
# compromise - lower means faster shutdown when quit but higher means lower
# cpu usage when running.

PAUSE_TIME = 0.4

SCHEMA_VERSION = 2  # version of our scheme, bump this if you change the following
SCHEMA = (
  'CREATE TABLE schema(version INTEGER)',
  'INSERT INTO schema(version) VALUES (%i)' % SCHEMA_VERSION,

  'CREATE TABLE metadata(relays_updated_at REAL)',
  'INSERT INTO metadata(relays_updated_at) VALUES (0.0)',

  'CREATE TABLE relays(fingerprint TEXT PRIMARY KEY, address TEXT, or_port INTEGER, nickname TEXT)',
  'CREATE INDEX addresses ON relays(address)',
)


try:
  uses_settings = stem.util.conf.uses_settings('nyx', os.path.join(BASE_DIR, 'settings'), lazy_load = False)
except IOError as exc:
  print("Unable to load nyx's internal configurations: %s" % exc)
  sys.exit(1)


def main():
  try:
    nyx.starter.main()
  except ImportError as exc:
    if exc.message == 'No module named stem':
      if stem.util.system.is_available('pip'):
        advice = ", try running 'sudo pip install stem'"
      elif stem.util.system.is_available('apt-get'):
        advice = ", try running 'sudo apt-get install python-stem'"
      else:
        advice = ', you can find it at https://stem.torproject.org/download.html'

      print('nyx requires stem' + advice)
    else:
      print('Unable to start nyx: %s' % exc)

    sys.exit(1)


def draw_loop():
  interface = nyx_interface()
  next_key = None  # use this as the next user input

  # Redrawing before starting daemons is important so our interface is rendered
  # right away and the 'top' positions are set for our panels.

  interface.redraw()

  for panel in interface:
    if isinstance(panel, nyx.panel.DaemonPanel):
      panel.start()

  stem.util.log.info('nyx started (initialization took %0.1f seconds)' % (time.time() - CONFIG['start_time']))

  while not interface._quit:
    if next_key:
      key, next_key = next_key, None
    else:
      key = nyx.curses.key_input(CONFIG['redraw_rate'])

    if key.match('right'):
      interface.set_page((interface.get_page() + 1) % interface.page_count())
    elif key.match('left'):
      interface.set_page((interface.get_page() - 1) % interface.page_count())
    elif key.match('p'):
      interface.set_paused(not interface.is_paused())
    elif key.match('m'):
      nyx.menu.show_menu()
    elif key.match('q'):
      if CONFIG['confirm_quit']:
        confirmation_key = show_message('Are you sure (q again to confirm)?', nyx.curses.BOLD, max_wait = 30)

        if not confirmation_key.match('q'):
          continue

      break
    elif key.match('x'):
      confirmation_key = show_message("This will reset Tor's internal state. Are you sure (x again to confirm)?", nyx.curses.BOLD, max_wait = 30)

      if confirmation_key.match('x'):
        try:
          tor_controller().signal(stem.Signal.RELOAD)
        except stem.ControllerError as exc:
          stem.util.log.error('Error detected when reloading tor: %s' % exc.strerror)
    elif key.match('h'):
      next_key = nyx.popups.show_help()
    elif not key.is_null():
      for panel in interface.page_panels():
        for keybinding in panel.key_handlers():
          keybinding.handle(key)

    interface.redraw(force = not key.is_null())


def nyx_interface():
  """
  Singleton controller for our interface.

  :returns: :class:`~nyx.Interface` controller
  """

  if NYX_INTERFACE is None:
    Interface()  # constructor sets NYX_INTERFACE

  return NYX_INTERFACE


def tor_controller():
  """
  Singleton for getting our tor controller connection.

  :returns: :class:`~stem.control.Controller` nyx is using
  """

  return TOR_CONTROLLER


def cache():
  """
  Provides the sqlite cache for application data.

  :returns: :class:`~nyx.cache.Cache` for our applicaion
  """

  global CACHE

  if CACHE is None:
    CACHE = Cache()

  return CACHE


def show_message(message = None, *attr, **kwargs):
  """
  Shows a message in our header.

  :param str message: message to be shown
  """

  return nyx_interface().header_panel().show_message(message, *attr, **kwargs)


def input_prompt(msg, initial_value = ''):
  """
  Prompts the user for input.

  :param str message: prompt for user input
  :param str initial_value: initial value of the prompt

  :returns: **str** with the user input, this is **None** if the prompt is
    canceled
  """

  header_panel = nyx_interface().header_panel()

  header_panel.show_message(msg)
  user_input = nyx.curses.str_input(len(msg), header_panel.get_height() - 1, initial_value)
  header_panel.show_message()

  return user_input


def init_controller(*args, **kwargs):
  """
  Sets the Controller used by nyx. This is a passthrough for Stem's
  :func:`~stem.connection.connect` function.

  :returns: :class:`~stem.control.Controller` nyx is using
  """

  global TOR_CONTROLLER
  TOR_CONTROLLER = stem.connection.connect(*args, **kwargs)
  return TOR_CONTROLLER


@uses_settings
def data_directory(filename, config):
  path = config.get('data_directory', '~/.nyx')

  if path == 'disabled':
    return None

  data_dir = os.path.expanduser(path)

  if not os.path.exists(data_dir):
    try:
      os.makedirs(data_dir)
    except OSError as exc:
      msg_runlevel = stem.util.log.INFO if 'Permission denied' in str(exc) else stem.util.log.NOTICE
      stem.util.log.log_once('nyx.data_directory_unavailable', msg_runlevel, 'Unable to create a data directory at %s (%s). This is fine, but caching is disabled meaning performance will be diminished a bit.' % (data_dir, exc))
      return None

  return os.path.join(data_dir, filename)


def expand_path(path):
  """
  Expands relative paths and include our chroot if one was set.

  :param str path: path to be expanded

  :returns: **str** with the expanded path
  """

  if path is None:
    return None

  try:
    tor_cwd = stem.util.system.cwd(tor_controller().get_pid(None)) if not os.path.isabs(path) else None
    return chroot() + stem.util.system.expand_path(path, tor_cwd)
  except IOError as exc:
    stem.util.log.info('Unable to expand a relative path (%s): %s' % (path, exc))
    return path


@uses_settings
def chroot(config):
  global CHROOT

  if CHROOT is None:
    # If the user provided us with a chroot then validate and normalize the
    # path.

    chroot = config.get('tor_chroot', '').strip().rstrip(os.path.sep)

    if chroot and not os.path.exists(chroot):
      stem.util.log.notice("The chroot path set in your config (%s) doesn't exist." % chroot)
      chroot = ''

    if not chroot and platform.system() == 'FreeBSD':
      controller = tor_controller()
      pid = controller.get_pid(None) if controller else stem.util.system.pid_by_name('tor')
      jail_chroot = stem.util.system.bsd_jail_path(pid) if pid else None

      if jail_chroot and os.path.exists(jail_chroot):
        stem.util.log.info('Adjusting paths to account for Tor running in a FreeBSD jail at: %s' % jail_chroot)
        chroot = jail_chroot

    CHROOT = chroot

  return CHROOT


def join(entries, joiner = ' ', size = None):
  """
  Joins a series of strings similar to str.join(), but only up to a given size.
  This returns an empty string if none of the entries will fit. For example...

    >>> join(['This', 'is', 'a', 'looooong', 'message'], size = 18)
    'This is a looooong'

    >>> join(['This', 'is', 'a', 'looooong', 'message'], size = 17)
    'This is a'

    >>> join(['This', 'is', 'a', 'looooong', 'message'], size = 2)
    ''

  :param list entries: strings to be joined
  :param str joiner: strings to join the entries with
  :param int size: maximum length the result can be, there's no length
    limitation if **None**

  :returns: **str** of the joined entries up to the given length
  """

  if size is None:
    return joiner.join(entries)

  result = ''

  for entry in entries:
    new_result = joiner.join((result, entry)) if result else entry

    if len(new_result) > size:
      break
    else:
      result = new_result

  return result


class Cache(object):
  """
  Cache for frequently needed information. This persists to disk if we can, and
  otherwise is an in-memory cache.
  """

  def __init__(self):
    self._conn_lock = threading.RLock()
    cache_path = nyx.data_directory('cache.sqlite')

    if cache_path and os.path.isfile(cache_path) and not os.access(cache_path, os.W_OK):
      stem.util.log.notice("Nyx's cache at %s is not writable by our user (%s). That's ok, but we'll have better performance if we can write to it." % (cache_path, getpass.getuser()))
      cache_path = None

    if cache_path:
      try:
        self._conn = sqlite3.connect(cache_path, check_same_thread = False)
        schema = self._query('SELECT version FROM schema').fetchone()[0]
      except:
        schema = None

      if schema == SCHEMA_VERSION:
        stem.util.log.info('Cache loaded from %s' % cache_path)
      else:
        if schema is None:
          stem.util.log.info('Cache at %s is missing a schema, clearing it.' % cache_path)
        else:
          stem.util.log.info('Cache at %s has schema version %s but the current version is %s, clearing it.' % (cache_path, schema, SCHEMA_VERSION))

        self._conn.close()
        os.remove(cache_path)
        self._conn = sqlite3.connect(cache_path, check_same_thread = False)

        for cmd in SCHEMA:
          self._conn.execute(cmd)
    else:
      stem.util.log.info('Unable to cache to disk. Using an in-memory cache instead.')
      self._conn = sqlite3.connect(':memory:', check_same_thread = False)

      for cmd in SCHEMA:
        self._conn.execute(cmd)

  @contextlib.contextmanager
  def write(self):
    """
    Provides a context in which we can modify the cache.

    :returns: :class:`~nyx.CacheWriter` that can modify the cache
    """

    with self._conn:
      yield CacheWriter(self)

  def relays_for_address(self, address):
    """
    Provides the relays running at a given location.

    :param str address: address to be checked

    :returns: **dict** of ORPorts to their fingerprint
    """

    result = {}

    for entry in self._query('SELECT or_port, fingerprint FROM relays WHERE address=?', address).fetchall():
      result[entry[0]] = entry[1]

    return result

  def relay_nickname(self, fingerprint, default = None):
    """
    Provides the nickname associated with the given relay.

    :param str fingerprint: relay to look up
    :param str default: response if no such relay exists

    :returns: **str** with the nickname ("Unnamed" if unset)
    """

    result = self._query('SELECT nickname FROM relays WHERE fingerprint=?', fingerprint).fetchone()
    return result[0] if result else default

  def relay_address(self, fingerprint, default = None):
    """
    Provides the (address, port) tuple where a relay is running.

    :param str fingerprint: fingerprint to be checked
    :param str default: response if no such relay exists

    :returns: **tuple** with a **str** address and **int** port
    """

    result = self._query('SELECT address, or_port FROM relays WHERE fingerprint=?', fingerprint).fetchone()
    return result if result else default

  def relays_updated_at(self):
    """
    Provides the unix timestamp when relay information was last updated.

    :returns: **float** with the unix timestamp when relay information was last
      updated, zero if it has never been set
    """

    return self._query('SELECT relays_updated_at FROM metadata').fetchone()[0]

  def _query(self, query, *param):
    """
    Performs a query on our cache.
    """

    with self._conn_lock:
      return self._conn.execute(query, param)


class CacheWriter(object):
  def __init__(self, cache):
    self._cache = cache

  def record_relay(self, fingerprint, address, or_port, nickname):
    """
    Records relay metadata.

    :param str fingerprint: relay fingerprint
    :param str address: ipv4 or ipv6 address
    :param int or_port: ORPort of the relay
    :param str nickname: relay nickname

    :raises: **ValueError** if provided data is malformed
    """

    if not stem.util.tor_tools.is_valid_fingerprint(fingerprint):
      raise ValueError("'%s' isn't a valid fingerprint" % fingerprint)
    elif not stem.util.tor_tools.is_valid_nickname(nickname):
      raise ValueError("'%s' isn't a valid nickname" % nickname)
    elif not stem.util.connection.is_valid_ipv4_address(address) and not stem.util.connection.is_valid_ipv6_address(address):
      raise ValueError("'%s' isn't a valid address" % address)
    elif not stem.util.connection.is_valid_port(or_port):
      raise ValueError("'%s' isn't a valid port" % or_port)

    self._cache._query('INSERT OR REPLACE INTO relays(fingerprint, address, or_port, nickname) VALUES (?,?,?,?)', fingerprint, address, or_port, nickname)
    self._cache._query('UPDATE metadata SET relays_updated_at=?', time.time())


class Interface(object):
  """
  Overall state of the nyx interface.
  """

  def __init__(self):
    global NYX_INTERFACE

    self._page = 0
    self._page_panels = []
    self._header_panel = None
    self._paused = False
    self._quit = False

    NYX_INTERFACE = self

    self._header_panel = nyx.panel.header.HeaderPanel()
    first_page_panels = []

    if CONFIG['show_graph']:
      first_page_panels.append(nyx.panel.graph.GraphPanel())

    if CONFIG['show_log']:
      first_page_panels.append(nyx.panel.log.LogPanel())

    if first_page_panels:
      self._page_panels.append(first_page_panels)

    if CONFIG['show_connections']:
      self._page_panels.append([nyx.panel.connection.ConnectionPanel()])

    if CONFIG['show_config']:
      self._page_panels.append([nyx.panel.config.ConfigPanel()])

    if CONFIG['show_torrc']:
      self._page_panels.append([nyx.panel.torrc.TorrcPanel()])

    if CONFIG['show_interpreter']:
      self._page_panels.append([nyx.panel.interpreter.InterpreterPanel()])

    visible_panels = self.page_panels()

    for panel in self:
      panel.set_visible(panel in visible_panels)

  def get_page(self):
    """
    Provides the page we're showing.

    :return: **int** of the page we're showing
    """

    return self._page

  def set_page(self, page_number):
    """
    Sets the selected page.

    :param int page_number: page to be shown

    :raises: **ValueError** if the page_number is invalid
    """

    if page_number < 0 or page_number >= self.page_count():
      raise ValueError('Invalid page number: %i' % page_number)

    if page_number != self._page:
      self._page = page_number
      self.header_panel().redraw()

      visible_panels = self.page_panels()

      for panel in self:
        panel.set_visible(panel in visible_panels)

  def page_count(self):
    """
    Provides the number of pages the interface has.

    :returns: **int** number of pages in the interface
    """

    return len(self._page_panels)

  def header_panel(self):
    """
    Provides our interface's header.

    :returns: :class:`~nyx.panel.header.HeaderPanel` of our interface
    """

    return self._header_panel

  def page_panels(self, page_number = None):
    """
    Provides panels belonging to a page, ordered top to bottom.

    :param int page_number: page to provide the panels of, current page if
      **None**

    :returns: **list** of panels on that page
    """

    if not self._page_panels:
      return [self._header_panel]  # all panels disabled

    page_number = self._page if page_number is None else page_number
    return [self._header_panel] + self._page_panels[page_number]

  def is_paused(self):
    """
    Checks if the interface is configured to be paused.

    :returns: **True** if the interface is paused, **False** otherwise
    """

    return self._paused

  def set_paused(self, is_pause):
    """
    Pauses or unpauses the interface.

    :param bool is_pause: suspends the interface if **True**, resumes it
      otherwise
    """

    if is_pause != self._paused:
      self._paused = is_pause

      for panel in self:
        panel.set_paused(is_pause)

      for panel in self.page_panels():
        panel.redraw()

  def redraw(self, force = False):
    """
    Renders our displayed content.
    """

    occupied = 0

    for panel in self.page_panels():
      panel.redraw(force = force, top = occupied)
      occupied += panel.get_height()

  def quit(self):
    """
    Quits our application.
    """

    self._quit = True

  def halt(self):
    """
    Stops curses panels in our interface.

    :returns: **threading.Thread** terminating daemons
    """

    def halt_panels():
      daemons = [panel for panel in self if isinstance(panel, nyx.panel.DaemonPanel)]

      for panel in daemons:
        panel.stop()

      for panel in daemons:
        panel.join()

    halt_thread = threading.Thread(target = halt_panels)
    halt_thread.start()
    return halt_thread

  def __iter__(self):
    yield self._header_panel

    for page in self._page_panels:
      for panel in page:
        yield panel


import nyx.curses
import nyx.menu
import nyx.panel
import nyx.panel.config
import nyx.panel.connection
import nyx.panel.graph
import nyx.panel.header
import nyx.panel.interpreter
import nyx.panel.log
import nyx.panel.torrc
import nyx.popups
import nyx.starter



# Copyright 2010-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

"""
Graphs of tor related statistics. For example...

Downloaded (0.0 B/sec):           Uploaded (0.0 B/sec):
  34                                30
                            *                                 *
                    **  *   *                          *      **
      *   *  *      ** **   **          ***  **       ** **   **
     *********      ******  ******     *********      ******  ******
   0 ************ ****************   0 ************ ****************
         25s  50   1m   1.6  2.0           25s  50   1m   1.6  2.0
"""

import copy
import functools
import threading
import time

import nyx.curses
import nyx.panel
import nyx.popups
import nyx.tracker

from nyx import nyx_interface, tor_controller, join, show_message
from nyx.curses import RED, GREEN, CYAN, BOLD, HIGHLIGHT
from nyx.menu import MenuItem, Submenu, RadioMenuItem, RadioGroup
from stem.control import EventType, Listener
from stem.util import conf, enum, log, str_tools, system

GraphStat = enum.Enum(('BANDWIDTH', 'bandwidth'), ('CONNECTIONS', 'connections'), ('SYSTEM_RESOURCES', 'resources'))
Interval = enum.Enum(('EACH_SECOND', 'each second'), ('FIVE_SECONDS', '5 seconds'), ('THIRTY_SECONDS', '30 seconds'), ('MINUTELY', 'minutely'), ('FIFTEEN_MINUTE', '15 minute'), ('THIRTY_MINUTE', '30 minute'), ('HOURLY', 'hourly'), ('DAILY', 'daily'))
Bounds = enum.Enum(('GLOBAL_MAX', 'global_max'), ('LOCAL_MAX', 'local_max'), ('TIGHT', 'tight'))

INTERVAL_SECONDS = {
  Interval.EACH_SECOND: 1,
  Interval.FIVE_SECONDS: 5,
  Interval.THIRTY_SECONDS: 30,
  Interval.MINUTELY: 60,
  Interval.FIFTEEN_MINUTE: 900,
  Interval.THIRTY_MINUTE: 1800,
  Interval.HOURLY: 3600,
  Interval.DAILY: 86400,
}

PRIMARY_COLOR, SECONDARY_COLOR = GREEN, CYAN

ACCOUNTING_RATE = 5
DEFAULT_CONTENT_HEIGHT = 4  # space needed for labeling above and below the graph
WIDE_LABELING_GRAPH_COL = 50  # minimum graph columns to use wide spacing for x-axis labels
TITLE_UPDATE_RATE = 30


def conf_handler(key, value):
  if key == 'graph_height':
    return max(1, value)
  elif key == 'max_graph_width':
    return max(1, value)
  elif key == 'graph_stat':
    if value != 'none' and value not in GraphStat:
      log.warn("'%s' isn't a valid graph type, options are: none, %s" % (CONFIG['graph_stat'], ', '.join(GraphStat)))
      return CONFIG['graph_stat']  # keep the default
  elif key == 'graph_interval':
    if value not in Interval:
      log.warn("'%s' isn't a valid graphing interval, options are: %s" % (value, ', '.join(Interval)))
      return CONFIG['graph_interval']  # keep the default
  elif key == 'graph_bound':
    if value not in Bounds:
      log.warn("'%s' isn't a valid graph bounds, options are: %s" % (value, ', '.join(Bounds)))
      return CONFIG['graph_bound']  # keep the default


CONFIG = conf.config_dict('nyx', {
  'attr.hibernate_color': {},
  'attr.graph.title': {},
  'attr.graph.header.primary': {},
  'attr.graph.header.secondary': {},
  'graph_bound': Bounds.LOCAL_MAX,
  'graph_height': 7,
  'graph_interval': Interval.EACH_SECOND,
  'graph_stat': GraphStat.BANDWIDTH,
  'max_graph_width': 300,  # we need some sort of max size so we know how much graph data to retain
  'show_accounting': True,
  'show_bits': False,
  'show_connections': True,
}, conf_handler)


def _bandwidth_title_stats():
  controller = tor_controller()

  stats = []
  bw_rate = controller.get_effective_rate(None)
  bw_burst = controller.get_effective_rate(None, burst = True)

  if bw_rate and bw_burst:
    bw_rate_label = _size_label(bw_rate)
    bw_burst_label = _size_label(bw_burst)

    # if both are using rounded values then strip off the '.0' decimal

    if '.0' in bw_rate_label and '.0' in bw_burst_label:
      bw_rate_label = bw_rate_label.replace('.0', '')
      bw_burst_label = bw_burst_label.replace('.0', '')

    stats.append('limit: %s/s' % bw_rate_label)
    stats.append('burst: %s/s' % bw_burst_label)

  my_server_descriptor = controller.get_server_descriptor(default = None)
  observed_bw = getattr(my_server_descriptor, 'observed_bandwidth', None)

  if observed_bw:
    stats.append('observed: %s/s' % _size_label(observed_bw))

  return stats


class GraphData(object):
  """
  Graphable statistical information.

  :var int latest_value: last value we recorded
  :var int total: sum of all values we've recorded
  :var int tick: number of events we've processed
  :var dict values: mapping of intervals to an array of samplings from newest to oldest
  """

  def __init__(self, clone = None, category = None, is_primary = True):
    if clone:
      self.latest_value = clone.latest_value
      self.total = clone.total
      self.tick = clone.tick
      self.values = copy.deepcopy(clone.values)

      self._category = category
      self._is_primary = clone._is_primary
      self._in_process_value = dict(clone._in_process_value)
      self._max_value = dict(clone._max_value)
    else:
      self.latest_value = 0
      self.total = 0
      self.tick = 0
      self.values = dict([(i, CONFIG['max_graph_width'] * [0]) for i in Interval])

      self._category = category
      self._is_primary = is_primary
      self._in_process_value = dict([(i, 0) for i in Interval])
      self._max_value = dict([(i, 0) for i in Interval])  # interval => maximum value it's had

  def average(self):
    return self.total / max(1, self.tick)

  def update(self, new_value):
    self.latest_value = new_value
    self.total += new_value
    self.tick += 1

    for interval in Interval:
      interval_seconds = INTERVAL_SECONDS[interval]
      self._in_process_value[interval] += new_value

      if self.tick % interval_seconds == 0:
        new_entry = self._in_process_value[interval] / interval_seconds
        self.values[interval] = [new_entry] + self.values[interval][:-1]
        self._max_value[interval] = max(self._max_value[interval], new_entry)
        self._in_process_value[interval] = 0

  def header(self, width):
    """
    Provides the description above a subgraph.

    :param int width: maximum length of the header

    :returns: **str** with our graph header
    """

    return self._category._header(width, self._is_primary)

  def bounds(self, bounds, interval, columns):
    """
    Range of values for the graph.

    :param Bounds bounds: boundary type for the range we want
    :param Interval interval: timing interval of the values
    :param int columns: number of values to take into account

    :returns: **tuple** of the form (min, max)
    """

    min_bound, max_bound = 0, 0
    values = self.values[interval][:columns]

    if bounds == Bounds.GLOBAL_MAX:
      max_bound = self._max_value[interval]
    elif columns > 0:
      max_bound = max(values)  # local maxima

    if bounds == Bounds.TIGHT and columns > 0:
      min_bound = min(values)

      # if the max = min pick zero so we still display something

      if min_bound == max_bound:
        min_bound = 0

    return min_bound, max_bound

  def y_axis_label(self, value):
    """
    Provides the label we should display on our y-axis.

    :param int value: value being shown on the y-axis

    :returns: **str** with our y-axis label
    """

    return self._category._y_axis_label(value, self._is_primary)


class GraphCategory(object):
  """
  Category for the graph. This maintains two subgraphs, updating them each
  second with updated stats.

  :var GraphData primary: first subgraph
  :var GraphData secondary: second subgraph
  :var float start_time: unix timestamp for when we started
  """

  def __init__(self, clone = None):
    if clone:
      self.primary = GraphData(clone.primary, category = self)
      self.secondary = GraphData(clone.secondary, category = self)
      self.start_time = clone.start_time
      self._title_stats = list(clone._title_stats)
      self._primary_header_stats = list(clone._primary_header_stats)
      self._secondary_header_stats = list(clone._secondary_header_stats)
    else:
      self.primary = GraphData(category = self, is_primary = True)
      self.secondary = GraphData(category = self, is_primary = False)
      self.start_time = time.time()
      self._title_stats = []
      self._primary_header_stats = []
      self._secondary_header_stats = []

  def stat_type(self):
    """
    Provides the GraphStat this graph is for.

    :returns: **GraphStat** of this graph
    """

    raise NotImplementedError('Should be implemented by subclasses')

  def title(self, width):
    """
    Provides a graph title that fits in the given width.

    :param int width: maximum length of the title

    :returns: **str** with our title
    """

    title = CONFIG['attr.graph.title'].get(self.stat_type(), '')
    title_stats = join(self._title_stats, ', ', width - len(title) - 4)
    return '%s (%s):' % (title, title_stats) if title_stats else title + ':'

  def bandwidth_event(self, event):
    """
    Called when it's time to process another event. All graphs use tor BW
    events to keep in sync with each other (this happens once per second).
    """

    pass

  def _header(self, width, is_primary):
    if is_primary:
      header = CONFIG['attr.graph.header.primary'].get(self.stat_type(), '')
      header_stats = self._primary_header_stats
    else:
      header = CONFIG['attr.graph.header.secondary'].get(self.stat_type(), '')
      header_stats = self._secondary_header_stats

    header_stats = join(header_stats, '', width - len(header) - 4).rstrip()
    return '%s (%s):' % (header, header_stats) if header_stats else '%s:' % header

  def _y_axis_label(self, value, is_primary):
    return str(value)


class BandwidthStats(GraphCategory):
  """
  Tracks tor's bandwidth usage.
  """

  def __init__(self, clone = None):
    GraphCategory.__init__(self, clone)
    self._title_last_updated = None

    if not clone:
      # fill in past bandwidth information

      controller = tor_controller()
      bw_entries, is_successful = controller.get_info('bw-event-cache', None), True

      if bw_entries:
        for entry in bw_entries.split():
          entry_comp = entry.split(',')

          if len(entry_comp) != 2 or not entry_comp[0].isdigit() or not entry_comp[1].isdigit():
            log.warn("Tor's 'GETINFO bw-event-cache' provided malformed output: %s" % bw_entries)
            is_successful = False
            break

          self.primary.update(int(entry_comp[0]))
          self.secondary.update(int(entry_comp[1]))

        if is_successful:
          log.info('Bandwidth graph has information for the last %s' % str_tools.time_label(len(bw_entries.split()), is_long = True))

      read_total = controller.get_info('traffic/read', None)
      write_total = controller.get_info('traffic/written', None)
      start_time = system.start_time(controller.get_pid(None))

      if read_total and write_total and start_time:
        self.primary.total = int(read_total)
        self.secondary.total = int(write_total)
        self.start_time = start_time

  def stat_type(self):
    return GraphStat.BANDWIDTH

  def _y_axis_label(self, value, is_primary):
    return _size_label(value, 0)

  def bandwidth_event(self, event):
    self.primary.update(event.read)
    self.secondary.update(event.written)

    self._primary_header_stats = [
      '%-14s' % ('%s/sec' % _size_label(self.primary.latest_value)),
      '- avg: %s/sec' % _size_label(self.primary.total / (time.time() - self.start_time)),
      ', total: %s' % _size_label(self.primary.total),
    ]

    self._secondary_header_stats = [
      '%-14s' % ('%s/sec' % _size_label(self.secondary.latest_value)),
      '- avg: %s/sec' % _size_label(self.secondary.total / (time.time() - self.start_time)),
      ', total: %s' % _size_label(self.secondary.total),
    ]

    if not self._title_last_updated or time.time() - self._title_last_updated > TITLE_UPDATE_RATE:
      self._title_stats = _bandwidth_title_stats()
      self._title_last_updated = time.time()


class ConnectionStats(GraphCategory):
  """
  Tracks number of inbound and outbound connections.
  """

  def stat_type(self):
    return GraphStat.CONNECTIONS

  def bandwidth_event(self, event):
    inbound_count, outbound_count = 0, 0

    controller = tor_controller()
    or_ports = controller.get_ports(Listener.OR, [])
    dir_ports = controller.get_ports(Listener.DIR, [])
    control_ports = controller.get_ports(Listener.CONTROL, [])

    for entry in nyx.tracker.get_connection_tracker().get_value():
      if entry.local_port in or_ports or entry.local_port in dir_ports:
        inbound_count += 1
      elif entry.local_port in control_ports:
        pass  # control connection
      else:
        outbound_count += 1

    self.primary.update(inbound_count)
    self.secondary.update(outbound_count)

    self._primary_header_stats = [str(self.primary.latest_value), ', avg: %i' % self.primary.average()]
    self._secondary_header_stats = [str(self.secondary.latest_value), ', avg: %i' % self.secondary.average()]


class ResourceStats(GraphCategory):
  """
  Tracks cpu and memory usage of the tor process.
  """

  def stat_type(self):
    return GraphStat.SYSTEM_RESOURCES

  def _y_axis_label(self, value, is_primary):
    return '%i%%' % value if is_primary else str_tools.size_label(value)

  def bandwidth_event(self, event):
    resources = nyx.tracker.get_resource_tracker().get_value()
    self.primary.update(resources.cpu_sample * 100)  # decimal percentage to whole numbers
    self.secondary.update(resources.memory_bytes)

    self._primary_header_stats = ['%0.1f%%' % self.primary.latest_value, ', avg: %0.1f%%' % self.primary.average()]
    self._secondary_header_stats = [str_tools.size_label(self.secondary.latest_value, 1), ', avg: %s' % str_tools.size_label(self.secondary.average(), 1)]


class GraphPanel(nyx.panel.Panel):
  """
  Panel displaying graphical information of GraphCategory instances.
  """

  def __init__(self):
    nyx.panel.Panel.__init__(self)

    self._displayed_stat = None if CONFIG['graph_stat'] == 'none' else CONFIG['graph_stat']
    self._update_interval = CONFIG['graph_interval']
    self._bounds_type = CONFIG['graph_bound']
    self._graph_height = CONFIG['graph_height']

    self._accounting_stats = None
    self._accounting_stats_paused = None

    self._stats = {
      GraphStat.BANDWIDTH: BandwidthStats(),
      GraphStat.SYSTEM_RESOURCES: ResourceStats(),
    }

    self._stats_lock = threading.RLock()
    self._stats_paused = None

    if CONFIG['show_connections']:
      self._stats[GraphStat.CONNECTIONS] = ConnectionStats()
    elif self._displayed_stat == GraphStat.CONNECTIONS:
      log.warn("The connection graph is unavailble when you set 'show_connections false'.")
      self._displayed_stat = GraphStat.BANDWIDTH

    controller = tor_controller()
    controller.add_event_listener(self._update_accounting, EventType.BW)
    controller.add_event_listener(self._update_stats, EventType.BW)
    controller.add_status_listener(lambda *args: self.redraw())

  def stat_options(self):
    return self._stats.keys()

  def get_height(self):
    """
    Provides the height of the content.
    """

    max_height = nyx.panel.Panel.get_height(self)

    if not self._displayed_stat:
      return 0

    height = DEFAULT_CONTENT_HEIGHT + self._graph_height
    accounting_stats = self._accounting_stats if not nyx_interface().is_paused() else self._accounting_stats_paused

    if self._displayed_stat == GraphStat.BANDWIDTH and accounting_stats:
      height += 3

    return min(max_height, height)

  def set_graph_height(self, new_graph_height):
    self._graph_height = max(1, new_graph_height)

  def _resize_graph(self):
    """
    Prompts for user input to resize the graph panel. Options include...

      * down arrow - grow graph
      * up arrow - shrink graph
      * enter / space - set size
    """

    with nyx.curses.CURSES_LOCK:
      try:
        while True:
          show_message('press the down/up to resize the graph, and enter when done', BOLD)
          key = nyx.curses.key_input()

          if key.match('down'):
            # don't grow the graph if it's already consuming the whole display
            # (plus an extra line for the graph/log gap)

            max_height = nyx.curses.screen_size().height - self.get_top()
            current_height = self.get_height()

            if current_height < max_height + 1:
              self.set_graph_height(self._graph_height + 1)
          elif key.match('up'):
            self.set_graph_height(self._graph_height - 1)
          elif key.is_selection():
            break

          nyx_interface().redraw()
      finally:
        show_message()

  def set_paused(self, is_pause):
    if is_pause:
      self._accounting_stats_paused = copy.copy(self._accounting_stats)
      self._stats_paused = dict([(key, type(self._stats[key])(self._stats[key])) for key in self._stats])

  def key_handlers(self):
    def _pick_stats():
      available_stats = sorted(self.stat_options())
      options = ['None'] + [stat.capitalize() for stat in available_stats]
      previous_selection = options[available_stats.index(self._displayed_stat) + 1] if self._displayed_stat else 'None'

      selection = nyx.popups.select_from_list('Graphed Stats:', options, previous_selection)
      self._displayed_stat = None if selection == 'None' else available_stats[options.index(selection) - 1]

    def _next_bounds():
      self._bounds_type = Bounds.next(self._bounds_type)
      self.redraw()

    def _pick_interval():
      self._update_interval = nyx.popups.select_from_list('Update Interval:', list(Interval), self._update_interval)
      self.redraw()

    return (
      nyx.panel.KeyHandler('g', 'resize graph', self._resize_graph),
      nyx.panel.KeyHandler('s', 'graphed stats', _pick_stats, self._displayed_stat if self._displayed_stat else 'none'),
      nyx.panel.KeyHandler('b', 'graph bounds', _next_bounds, self._bounds_type.replace('_', ' ')),
      nyx.panel.KeyHandler('i', 'graph update interval', _pick_interval, self._update_interval),
    )

  def submenu(self):
    """
    Submenu consisting of...

      [X] <Stat 1>
      [ ] <Stat 2>
      [ ] <Stat 2>
          Resize...
          Interval (Submenu)
          Bounds (Submenu)
    """

    stat_group = RadioGroup(functools.partial(setattr, self, '_displayed_stat'), self._displayed_stat)
    interval_group = RadioGroup(functools.partial(setattr, self, '_update_interval'), self._update_interval)
    bounds_group = RadioGroup(functools.partial(setattr, self, '_bounds_type'), self._bounds_type)

    return Submenu('Graph', [
      RadioMenuItem('None', stat_group, None),
      [RadioMenuItem(str_tools._to_camel_case(opt, divider = ' '), stat_group, opt) for opt in sorted(self.stat_options())],
      MenuItem('Resize...', self._resize_graph),
      Submenu('Interval', [RadioMenuItem(opt, interval_group, opt) for opt in Interval]),
      Submenu('Bounds', [RadioMenuItem(opt, bounds_group, opt) for opt in Bounds]),
    ])

  def _draw(self, subwindow):
    if not self._displayed_stat:
      return

    if not nyx_interface().is_paused():
      stat = self._stats[self._displayed_stat]
      accounting_stats = self._accounting_stats
    else:
      if not self._stats_paused:
        return  # when first paused concurrency could mean this isn't set yet

      stat = self._stats_paused[self._displayed_stat]
      accounting_stats = self._accounting_stats_paused

    with self._stats_lock:
      subgraph_height = self._graph_height + 2  # graph rows + header + x-axis label
      subgraph_width = min(subwindow.width // 2, CONFIG['max_graph_width'])
      interval, bounds_type = self._update_interval, self._bounds_type

      subwindow.addstr(0, 0, stat.title(subwindow.width), HIGHLIGHT)

      _draw_subgraph(subwindow, stat.primary, 0, subgraph_width, subgraph_height, bounds_type, interval, PRIMARY_COLOR)
      _draw_subgraph(subwindow, stat.secondary, subgraph_width, subgraph_width, subgraph_height, bounds_type, interval, SECONDARY_COLOR)

      if stat.stat_type() == GraphStat.BANDWIDTH and accounting_stats:
        _draw_accounting_stats(subwindow, DEFAULT_CONTENT_HEIGHT + subgraph_height - 2, accounting_stats)

  def _update_accounting(self, event):
    if not CONFIG['show_accounting']:
      self._accounting_stats = None
    elif not self._accounting_stats or time.time() - self._accounting_stats.retrieved >= ACCOUNTING_RATE:
      old_accounting_stats = self._accounting_stats
      self._accounting_stats = tor_controller().get_accounting_stats(None)

      if not nyx_interface().is_paused():
        # if we either added or removed accounting info then redraw the whole
        # screen to account for resizing

        if bool(old_accounting_stats) != bool(self._accounting_stats):
          nyx_interface().redraw()

  def _update_stats(self, event):
    with self._stats_lock:
      for stat in self._stats.values():
        stat.bandwidth_event(event)

    if self._displayed_stat:
      param = self._stats[self._displayed_stat]
      update_rate = INTERVAL_SECONDS[self._update_interval]

      if param.primary.tick % update_rate == 0:
        self.redraw()


def _draw_subgraph(subwindow, data, x, width, height, bounds_type, interval, color, fill_char = ' '):
  """
  Renders subgraph including its title, labeled axis, and content.
  """

  columns = width - 8  # y-axis labels can be at most six characters wide with a space on either side
  min_bound, max_bound = data.bounds(bounds_type, interval, columns)

  x_axis_labels = _x_axis_labels(interval, columns)
  y_axis_labels = _y_axis_labels(height, data, min_bound, max_bound)

  x_axis_offset = max([len(label) for label in y_axis_labels.values()])
  columns = max(columns, width - x_axis_offset - 2)

  subwindow.addstr(x, 1, data.header(width), color, BOLD)

  for x_offset, label in x_axis_labels.items():
    subwindow.addstr(x + x_offset + x_axis_offset, height, label, color)

  for y, label in y_axis_labels.items():
    subwindow.addstr(x, y, label, color)

  for col in range(columns):
    column_count = int(data.values[interval][col]) - min_bound
    column_height = int(min(height - 2, (height - 2) * column_count / (max(1, max_bound) - min_bound)))
    subwindow.vline(x + col + x_axis_offset + 1, height - column_height, column_height, color, HIGHLIGHT, char = fill_char)


def _x_axis_labels(interval, columns):
  """
  Provides the labels for the x-axis. We include the units for only its first
  value, then bump the precision for subsequent units. For example...

    10s, 20, 30, 40, 50, 1m, 1.1, 1.3, 1.5
  """

  x_axis_labels = {}

  interval_sec = INTERVAL_SECONDS[interval]
  interval_spacing = 10 if columns >= WIDE_LABELING_GRAPH_COL else 5
  previous_units, decimal_precision = None, 0

  for i in range((columns - 4) // interval_spacing):
    x = (i + 1) * interval_spacing
    time_label = str_tools.time_label(x * interval_sec, decimal_precision)

    if not previous_units:
      previous_units = time_label[-1]
    elif previous_units != time_label[-1]:
      previous_units = time_label[-1]
      decimal_precision = 1  # raised precision for future measurements
    else:
      time_label = time_label[:-1]  # strip units since already provided

    x_axis_labels[x] = time_label

  return x_axis_labels


def _y_axis_labels(subgraph_height, data, min_bound, max_bound):
  """
  Provides the labels for the y-axis. This is a mapping of the position it
  should be drawn at to its text.
  """

  y_axis_labels = {
    2: data.y_axis_label(max_bound),
    subgraph_height - 1: data.y_axis_label(min_bound),
  }

  ticks = (subgraph_height - 5) // 2

  for i in range(ticks):
    row = subgraph_height - (2 * i) - 5

    if subgraph_height % 2 == 0 and i >= (ticks // 2):
      row -= 1  # make extra gap be in the middle when we're an even size

    val = (max_bound - min_bound) * (subgraph_height - row - 3) // (subgraph_height - 3)

    if val not in (min_bound, max_bound):
      y_axis_labels[row + 2] = data.y_axis_label(val)

  return y_axis_labels


def _draw_accounting_stats(subwindow, y, accounting):
  if tor_controller().is_alive():
    hibernate_color = CONFIG['attr.hibernate_color'].get(accounting.status, RED)

    x = subwindow.addstr(0, y, 'Accounting (', BOLD)
    x = subwindow.addstr(x, y, accounting.status, BOLD, hibernate_color)
    x = subwindow.addstr(x, y, ')', BOLD)

    subwindow.addstr(35, y, 'Time to reset: %s' % str_tools.short_time_label(accounting.time_until_reset))

    subwindow.addstr(2, y + 1, '%s / %s' % (_size_label(accounting.read_bytes), _size_label(accounting.read_limit)), PRIMARY_COLOR)
    subwindow.addstr(37, y + 1, '%s / %s' % (_size_label(accounting.written_bytes), _size_label(accounting.write_limit)), SECONDARY_COLOR)
  else:
    subwindow.addstr(0, y, 'Accounting:', BOLD)
    subwindow.addstr(12, y, 'Connection Closed...')


def _size_label(byte_count, decimal = 1):
  """
  Alias for str_tools.size_label() that accounts for if the user prefers bits
  or bytes.
  """

  return str_tools.size_label(byte_count, decimal, is_bytes = not CONFIG['show_bits'], round = True)



# Copyright 2009-2019, Damian Johnson and The Tor Project
# See LICENSE for licensing information

"""
Top panel for every page, containing basic system and tor related information.
This expands the information it presents to two columns if there's room
available.
"""

import os
import time

import stem
import stem.control
import stem.util.proc
import stem.util.str_tools
import stem.util.system

import nyx
import nyx.curses
import nyx.panel
import nyx.popups
import nyx.tracker

from stem.util import conf, log
from nyx import nyx_interface, tor_controller

from nyx.curses import RED, GREEN, YELLOW, CYAN, WHITE, BOLD, HIGHLIGHT

MIN_DUAL_COL_WIDTH = 141  # minimum width where we'll show two columns
SHOW_FD_THRESHOLD = 60  # show file descriptor usage if usage is over this percentage
UPDATE_RATE = 5  # rate in seconds at which we refresh

CONFIG = conf.config_dict('nyx', {
  'attr.flag_colors': {},
  'attr.version_status_colors': {},
})


class HeaderPanel(nyx.panel.DaemonPanel):
  """
  Top area containing tor settings and system information.
  """

  def __init__(self):
    nyx.panel.DaemonPanel.__init__(self, UPDATE_RATE)
    self._vals = Sampling.create()

    self._last_width = nyx.curses.screen_size().width
    self._reported_inactive = False
    self._pause_time = 0

    self._message = None
    self._message_attr = []

    tor_controller().add_status_listener(self._reset_listener)

  def show_message(self, message = None, *attr, **kwargs):
    """
    Sets the message displayed at the bottom of the header. If not called with
    anything it clears the override.

    :param str message: message to be displayed
    :param list attr: text attributes to apply
    :param int max_wait: seconds to wait for user input, no limit if **None**

    :returns: :class:`~nyx.curses.KeyInput` user pressed if provided a
      **max_wait**, **None** otherwise or if prompt was canceled
    """

    self._message = message
    self._message_attr = attr
    self.redraw()

    if 'max_wait' in kwargs:
      user_input = nyx.curses.key_input(kwargs['max_wait'])
      self.show_message()  # clear override
      return user_input

  def is_wide(self):
    """
    True if we should show two columns of information, False otherwise.
    """

    return self._last_width >= MIN_DUAL_COL_WIDTH

  def get_height(self):
    """
    Provides the height of the content, which is dynamically determined by the
    panel's maximum width.
    """

    max_height = nyx.panel.DaemonPanel.get_height(self)

    if self._vals.is_relay:
      return min(max_height, 5 if self.is_wide() else 7)
    else:
      return min(max_height, 4 if self.is_wide() else 5)

  def send_newnym(self):
    """
    Requests a new identity and provides a visual queue.
    """

    controller = tor_controller()

    if not controller.is_newnym_available():
      return

    controller.signal(stem.Signal.NEWNYM)

    # If we're wide then the newnym label in this panel will give an
    # indication that the signal was sent. Otherwise use a msg.

    if not self.is_wide():
      self.show_message('Requesting a new identity', HIGHLIGHT, max_wait = 1)

  def set_paused(self, is_pause):
    if is_pause:
      self._pause_time = time.time()

  def key_handlers(self):
    def _reconnect():
      if self._vals.is_connected:
        return

      controller = tor_controller()
      self.show_message('Reconnecting...', HIGHLIGHT)

      try:
        try:
          controller.reconnect(chroot_path = nyx.chroot())
        except stem.connection.MissingPassword:
          password = nyx.input_prompt('Controller Password: ')

          if password:
            controller.authenticate(password)

        log.notice("Reconnected to Tor's control port")
        self.show_message('Tor reconnected', HIGHLIGHT, max_wait = 1)
      except Exception as exc:
        self.show_message('Unable to reconnect (%s)' % exc, HIGHLIGHT, max_wait = 3)
        controller.close()

    return (
      nyx.panel.KeyHandler('n', action = self.send_newnym),
      nyx.panel.KeyHandler('r', action = _reconnect),
    )

  def _draw(self, subwindow):
    vals = self._vals  # local reference to avoid concurrency concerns
    self._last_width = subwindow.width
    is_wide = self.is_wide()

    # space available for content

    interface = nyx_interface()
    left_width = max(subwindow.width // 2, 77) if is_wide else subwindow.width
    right_width = subwindow.width - left_width

    _draw_platform_section(subwindow, 0, 0, left_width, vals)

    if vals.is_connected:
      _draw_ports_section(subwindow, 0, 1, left_width, vals)
    else:
      _draw_disconnected(subwindow, 0, 1, vals.last_heartbeat)

    if is_wide:
      _draw_resource_usage(subwindow, left_width, 0, right_width, vals, self._pause_time)

      if vals.is_relay:
        _draw_fingerprint_and_fd_usage(subwindow, left_width, 1, right_width, vals)
        _draw_flags(subwindow, 0, 2, vals.flags)
        _draw_exit_policy(subwindow, left_width, 2, vals.exit_policy)
      elif vals.is_connected:
        _draw_newnym_option(subwindow, left_width, 1, vals.newnym_wait)
    else:
      _draw_resource_usage(subwindow, 0, 2, left_width, vals, self._pause_time)

      if vals.is_relay:
        _draw_fingerprint_and_fd_usage(subwindow, 0, 3, left_width, vals)
        _draw_flags(subwindow, 0, 4, vals.flags)

    _draw_status(subwindow, 0, self.get_height() - 1, interface.is_paused(), self._message, *self._message_attr)

  def _reset_listener(self, controller, event_type, _):
    self._update()

    if event_type == stem.control.State.CLOSED:
      log.notice('Tor control port closed')

  def _update(self):
    self._vals = Sampling.create(self._vals)

    if self._vals.fd_used and self._vals.fd_limit != -1:
      fd_percent = 100 * self._vals.fd_used // self._vals.fd_limit

      if fd_percent >= 90:
        log_msg = "Tor's file descriptor usage is at %s%%. If you run out Tor will be unable to continue functioning." % fd_percent
        log.log_once('fd_used_at_ninety_percent', log.WARN, log_msg)
        log.DEDUPLICATION_MESSAGE_IDS.add('fd_used_at_sixty_percent')
      elif fd_percent >= 60:
        log_msg = "Tor's file descriptor usage is at %s%%." % fd_percent
        log.log_once('fd_used_at_sixty_percent', log.NOTICE, log_msg)

    if self._vals.is_connected:
      if not self._reported_inactive and (time.time() - self._vals.last_heartbeat) >= 10:
        self._reported_inactive = True
        log.notice('Relay unresponsive (last heartbeat: %s)' % time.ctime(self._vals.last_heartbeat))
      elif self._reported_inactive and (time.time() - self._vals.last_heartbeat) < 10:
        self._reported_inactive = False
        log.notice('Relay resumed')

    self.redraw()


class Sampling(object):
  def __init__(self, **attr):
    self._attr = attr

    for key, value in attr.items():
      setattr(self, key, value)

  @staticmethod
  def create(last_sampling = None):
    controller = tor_controller()
    retrieved = time.time()

    pid = controller.get_pid('')
    tor_resources = nyx.tracker.get_resource_tracker().get_value()
    nyx_total_cpu_time = sum(os.times()[:3], stem.util.system.SYSTEM_CALL_TIME)

    or_listeners = controller.get_listeners(stem.control.Listener.OR, [])
    control_listeners = controller.get_listeners(stem.control.Listener.CONTROL, [])
    my_router_status_entry = nyx.tracker.get_consensus_tracker().my_router_status_entry()

    if controller.get_conf('HashedControlPassword', None):
      auth_type = 'password'
    elif controller.get_conf('CookieAuthentication', None) == '1':
      auth_type = 'cookie'
    else:
      auth_type = 'open'

    try:
      fd_used = stem.util.proc.file_descriptors_used(pid)
    except IOError:
      fd_used = None

    if last_sampling:
      nyx_cpu_delta = nyx_total_cpu_time - last_sampling.nyx_total_cpu_time
      nyx_time_delta = retrieved - last_sampling.retrieved
      nyx_cpu = nyx_cpu_delta / nyx_time_delta
    else:
      nyx_cpu = 0.0

    attr = {
      'retrieved': retrieved,
      'is_connected': controller.is_alive(),
      'connection_time': controller.connection_time(),
      'last_heartbeat': controller.get_latest_heartbeat(),

      'fingerprint': controller.get_info('fingerprint', 'Unknown'),
      'nickname': controller.get_conf('Nickname', ''),
      'newnym_wait': controller.get_newnym_wait(),
      'exit_policy': controller.get_exit_policy(None),
      'flags': getattr(my_router_status_entry, 'flags', []),

      'version': str(controller.get_version('Unknown')).split()[0],
      'version_status': controller.get_info('status/version/current', 'Unknown'),

      'address': or_listeners[0][0] if (or_listeners and or_listeners[0][0] != '0.0.0.0') else controller.get_info('address', 'Unknown'),
      'or_port': or_listeners[0][1] if or_listeners else '',
      'dir_port': controller.get_conf('DirPort', '0'),
      'control_port': str(control_listeners[0][1]) if control_listeners else None,
      'socket_path': controller.get_conf('ControlSocket', None),
      'is_relay': bool(or_listeners),

      'auth_type': auth_type,
      'pid': pid,
      'start_time': stem.util.system.start_time(pid),
      'fd_limit': int(controller.get_info('process/descriptor-limit', '-1')),
      'fd_used': fd_used,

      'nyx_total_cpu_time': nyx_total_cpu_time,
      'tor_cpu': '%0.1f' % (100 * tor_resources.cpu_sample),
      'nyx_cpu': '%0.1f' % (100 * nyx_cpu),
      'memory': stem.util.str_tools.size_label(tor_resources.memory_bytes) if tor_resources.memory_bytes > 0 else 0,
      'memory_percent': '%0.1f' % (100 * tor_resources.memory_percent),

      'hostname': os.uname()[1],
      'platform': '%s %s' % (os.uname()[0], os.uname()[2]),  # [platform name] [version]
    }

    return Sampling(**attr)

  def format(self, message, crop_width = None):
    formatted_msg = message.format(**self._attr)

    if crop_width is not None:
      formatted_msg = stem.util.str_tools.crop(formatted_msg, crop_width)

    return formatted_msg


def _draw_platform_section(subwindow, x, y, width, vals):
  """
  Section providing the user's hostname, platform, and version information...

    nyx - odin (Linux 3.5.0-52-generic)        Tor 0.2.5.1-alpha-dev (unrecommended)
    |------ platform (40 characters) ------|   |----------- tor version -----------|
  """

  initial_x, space_left = x, min(width, 40)

  x = subwindow.addstr(x, y, vals.format('nyx - {hostname}', space_left))
  space_left -= x - initial_x

  if space_left >= 10:
    subwindow.addstr(x, y, ' (%s)' % vals.format('{platform}', space_left - 3))

  x, space_left = initial_x + 43, width - 43

  if vals.version != 'Unknown' and space_left >= 10:
    x = subwindow.addstr(x, y, vals.format('Tor {version}', space_left))
    space_left -= x - 43 - initial_x

    if space_left >= 7 + len(vals.version_status):
      version_color = CONFIG['attr.version_status_colors'].get(vals.version_status, WHITE)

      x = subwindow.addstr(x, y, ' (')
      x = subwindow.addstr(x, y, vals.version_status, version_color)
      subwindow.addstr(x, y, ')')


def _draw_ports_section(subwindow, x, y, width, vals):
  """
  Section providing our nickname, address, and port information...

    Unnamed - 0.0.0.0:7000, Control Port (cookie): 9051
  """

  if not vals.is_relay:
    x = subwindow.addstr(x, y, 'Relaying Disabled', CYAN)
  else:
    x = subwindow.addstr(x, y, vals.format('{nickname} - {address}:{or_port}'))

    if vals.dir_port != '0':
      x = subwindow.addstr(x, y, vals.format(', Dir Port: {dir_port}'))

  if vals.control_port:
    if width >= x + 19 + len(vals.control_port) + len(vals.auth_type):
      auth_color = RED if vals.auth_type == 'open' else GREEN

      x = subwindow.addstr(x, y, ', Control Port (')
      x = subwindow.addstr(x, y, vals.auth_type, auth_color)
      subwindow.addstr(x, y, vals.format('): {control_port}'))
    else:
      subwindow.addstr(x, y, vals.format(', Control Port: {control_port}'))
  elif vals.socket_path:
    subwindow.addstr(x, y, vals.format(', Control Socket: {socket_path}'))


def _draw_disconnected(subwindow, x, y, last_heartbeat):
  """
  Message indicating that tor is disconnected...

    Tor Disconnected (15:21 07/13/2014, press r to reconnect)
  """

  x = subwindow.addstr(x, y, 'Tor Disconnected', RED, BOLD)
  last_heartbeat_str = time.strftime('%H:%M %m/%d/%Y', time.localtime(last_heartbeat))
  subwindow.addstr(x, y, ' (%s, press r to reconnect)' % last_heartbeat_str)


def _draw_resource_usage(subwindow, x, y, width, vals, pause_time):
  """
  System resource usage of the tor process...

    cpu: 0.0% tor, 1.0% nyx    mem: 0 (0.0%)       pid: 16329  uptime: 12-20:42:07
  """

  if vals.start_time:
    if not vals.is_connected:
      now = vals.connection_time
    elif pause_time:
      now = pause_time
    else:
      now = time.time()

    uptime = stem.util.str_tools.short_time_label(max(0, now - vals.start_time))
  else:
    uptime = ''

  sys_fields = (
    (0, vals.format('cpu: {tor_cpu}% tor, {nyx_cpu}% nyx')),
    (27, vals.format('mem: {memory} ({memory_percent}%)')),
    (47, vals.format('pid: {pid}')),
    (59, 'uptime: %s' % uptime),
  )

  for (start, label) in sys_fields:
    if width >= start + len(label):
      subwindow.addstr(x + start, y, label)
    else:
      break


def _draw_fingerprint_and_fd_usage(subwindow, x, y, width, vals):
  """
  Presents our fingerprint, and our file descriptor usage if we're running
  out...

    fingerprint: 1A94D1A794FCB2F8B6CBC179EF8FDD4008A98D3B, file desc: 900 / 1000 (90%)
  """

  initial_x, space_left = x, width

  x = subwindow.addstr(x, y, vals.format('fingerprint: {fingerprint}', width))
  space_left -= x - initial_x

  if space_left >= 30 and vals.fd_used and vals.fd_limit != -1:
    fd_percent = 100 * vals.fd_used / vals.fd_limit

    if fd_percent >= SHOW_FD_THRESHOLD:
      if fd_percent >= 95:
        percentage_format = (RED, BOLD)
      elif fd_percent >= 90:
        percentage_format = (RED,)
      elif fd_percent >= 60:
        percentage_format = (YELLOW,)
      else:
        percentage_format = ()

      x = subwindow.addstr(x, y, ', file descriptors' if space_left >= 37 else ', file desc')
      x = subwindow.addstr(x, y, vals.format(': {fd_used} / {fd_limit} ('))
      x = subwindow.addstr(x, y, '%i%%' % fd_percent, *percentage_format)
      subwindow.addstr(x, y, ')')


def _draw_flags(subwindow, x, y, flags):
  """
  Presents flags held by our relay...

    flags: Running, Valid
  """

  x = subwindow.addstr(x, y, 'flags: ')

  if flags:
    for i, flag in enumerate(flags):
      flag_color = CONFIG['attr.flag_colors'].get(flag, WHITE)
      x = subwindow.addstr(x, y, flag, flag_color, BOLD)

      if i < len(flags) - 1:
        x = subwindow.addstr(x, y, ', ')
  else:
    subwindow.addstr(x, y, 'none', CYAN, BOLD)


def _draw_exit_policy(subwindow, x, y, exit_policy):
  """
  Presents our exit policy...

    exit policy: reject *:*
  """

  x = subwindow.addstr(x, y, 'exit policy: ')

  if not exit_policy:
    return

  rules = list(exit_policy.strip_private().strip_default())

  for i, rule in enumerate(rules):
    policy_color = GREEN if rule.is_accept else RED
    x = subwindow.addstr(x, y, str(rule), policy_color, BOLD)

    if i < len(rules) - 1:
      x = subwindow.addstr(x, y, ', ')

  if exit_policy.has_default():
    if rules:
      x = subwindow.addstr(x, y, ', ')

    subwindow.addstr(x, y, '<default>', CYAN, BOLD)


def _draw_newnym_option(subwindow, x, y, newnym_wait):
  """
  Provide a notice for requiesting a new identity, and time until it's next
  available if in the process of building circuits.
  """

  if newnym_wait == 0:
    subwindow.addstr(x, y, "press 'n' for a new identity")
  else:
    plural = 's' if newnym_wait > 1 else ''
    subwindow.addstr(x, y, 'building circuits, available again in %i second%s' % (newnym_wait, plural))


def _draw_status(subwindow, x, y, is_paused, message, *attr):
  """
  Provides general usage information or a custom message.
  """

  if message:
    subwindow.addstr(x, y, message, *attr)
  elif not is_paused:
    interface = nyx_interface()
    subwindow.addstr(x, y, 'page %i / %i - m: menu, p: pause, h: page help, q: quit' % (interface.get_page() + 1, interface.page_count()))
  else:
    subwindow.addstr(x, y, 'Paused', HIGHLIGHT)
